/*
 * tunrouter is a glue software that binds to a TAP interface and moves packets
 * back and forth to SLIRP, taking care of virtual ARP resolution in the
 * process.
 * 
 * I created it specifically for making DOSemu networking easier, but it might
 * have other interesting use cases, too.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation,  either version 3 of the License,  or (at your option)  any later
 * version.
 * 
 * This program is  distributed in the hope that it will be useful,  but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * Copyright (C) Mateusz Viste 2013. All rights reserved.
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/if.h>
#include <linux/if_tun.h>
#include <sys/select.h>

#include "slirp.c"
#include "arp.c"
#include "netparse.c"
#include "forgehdr.c"
#include "processpkt.c"


#define PVER "0.51"



/* allocate a TUN/TAP interface. *dev is the name of the interface (can be "\0", then a name will be computed), flags is IFF_TUN or IFF_TAP.
 Returns a valid socket, or a negative value on error. */
int tun_alloc(char *dev, int flags) {
  struct ifreq ifr;
  int fd, err;
  char *clonedev = "/dev/net/tun";

  /* open the clone device */
  if ((fd = open(clonedev, O_RDWR)) < 0) {
    return(fd);
  }

  /* preparation of the struct ifr, of type "struct ifreq" */
  memset(&ifr, 0, sizeof(ifr));
  ifr.ifr_flags = flags;   /* IFF_TUN or IFF_TAP, plus maybe IFF_NO_PI */

  if (dev[0] != 0) {
    /* if a device name was specified, put it in the structure; otherwise,
     * the kernel will try to allocate the "next" device of the
     * specified type */
    strncpy(ifr.ifr_name, dev, IFNAMSIZ);
  }

  /* try to create the device */
  if ((err = ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0 ) {
    close(fd);
    return(err);
  }

  /* if the operation was successful, write back the name of the
   * interface to the variable "dev", so the caller can know
   * it. Note that the caller MUST reserve space in *dev (see calling
   * code below) */
  strcpy(dev, ifr.ifr_name);

  /* this is the special file descriptor that the caller will use to talk
   * with the virtual interface */
  return(fd);
}






int main(int argc, char **argv) {
  uint8_t mymac[] = {0xF0,0xDE,0xF1,0x29,0x07,0x85};
  uint32_t myip[] = {(10 << 24) | (0 << 16) | (2 << 8) | 1,  /* me as the gateway */
                     (10 << 24) | (0 << 16) | (2 << 8) | 2,  /* me as the host */
                     (10 << 24) | (0 << 16) | (2 << 8) | 3,  /* me as the dns server */
                     0};                                     /* last entry is marked as '0' */
  int mymask = 24;
  char tap_name[IFNAMSIZ];
  int tapfd, bufflen;
  #define buffmaxlen 4096
  uint8_t buff[buffmaxlen];
  int highestfd;
  int slirpfd[2];
  int x;
  char *slirpexec;
  uint32_t mynetmask;
  fd_set rfds;
  struct timeval tv;
  struct arptabletype *arptable = NULL;

  if (argc != 3) {
    puts("taprouter " PVER " Copyright (C) Mateusz Viste 2013.");
    puts("usage: taprouter ifname slirpfile");
    puts("example: taprouter tap0 /usr/bin/slirp");
    return(1);
  }

  /* compute the CIDR netmask from bitlen mask (for faster computation later) */
  switch (mymask) {
    case 8:
      mynetmask = 0xFF000000;
      break;
    case 16:
      mynetmask = 0xFFFF0000;
      break;
    case 24:
      mynetmask = 0xFFFFFF00;
      break;
    default:
      printf("Invalid netmask configure (/%d). Please use a /8, /16 or /24 mask.\n", mymask);
      return(1);
      break;
  }

  slirpexec = argv[2];

  strcpy(tap_name, argv[1]);
  tapfd = tun_alloc(tap_name, IFF_TAP|IFF_NO_PI);  /* attach the process to the tap interface, and tell it that we want a TAP interface without any extra metada (IFF_NO_PI tells that we want pure eth traffic) */

  if (tapfd < 0) {
    printf("Error while trying to allocate the tap interface '%s'. Sorry.\n", tap_name);
    if (errno == EPERM) printf("You seem to lack permissions. Are you the owner of the tap interface?\n");
    return(1);
  }
  puts("bind to tap interface: sucess.");

  if (slirp_open(slirpexec, slirpfd) != 0) {
    printf("Error: failed to invoke '%s'.\n", slirpexec);
    return(1);
  }
  puts("open SLIRP channel: success.");
  
  printf("\ntaprouter is running attached to '%s'.\n\nNetwork configuration to use on your guest:\n IP address: 10.0.2.15/24\n Gateway: 10.0.2.1\n DNS server: 10.0.2.3\n", tap_name);

  /* add all my IP addresses to the ARP cache, to avoid distributing them later to clients via DHCP */
  for (x = 0; myip[x] != 0; x++) {
    arptable = arp_learn(arptable, mymac, myip[x]);
  }

  for (;;) {

    FD_ZERO(&rfds);

    /* add all sockets to monitor into the list */
    FD_SET(tapfd, &rfds);
    FD_SET(slirpfd[0], &rfds);
    highestfd = tapfd;
    if (slirpfd[0] > highestfd) highestfd = slirpfd[0];

    /* Wait up to five seconds. */
    tv.tv_sec = 5;
    tv.tv_usec = 0;

    /* wait for something to happen */
    select(highestfd+1, &rfds, NULL, NULL, &tv);

    /* is there something on the tap iface? */
    if (FD_ISSET(tapfd, &rfds) != 0) {
      bufflen = read(tapfd, buff, buffmaxlen);
      if (bufflen < 0) {
        printf("Reading from interface '%s' failed. Sorry.\n", tap_name);
        close(tapfd);
        return(1);
      }
      processpkt(buff, bufflen, myip, mynetmask, mymac, slirpfd, &arptable, tapfd);
    }

    /* is there something on the slirp iface? */
    if (FD_ISSET(slirpfd[0], &rfds) != 0) {
      uint8_t *buffptr;
      buffptr = buff + 128; /* leaving some place before data, for headers */
      bufflen = slirp_read(buffptr, slirpfd);
      /* printf("Got %d bytes on the slirp iface.\n", bufflen); */
      if (bufflen > 0) { /* ignore 0-len packets (these are generated by SLIP) */
        uint8_t *dstmac;
        uint8_t *ipsrcarr, *ipdstarr;
        uint32_t ipdst;
        int ipprotocol, dscp, ttl, id, fragoffset, morefragsflag;
        uint8_t ethbroadcast[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
        /* analyze the IP header */
        parse_ipv4(buffptr, bufflen, &ipsrcarr, &ipdstarr, &ipprotocol, &dscp, &ttl, &id, &fragoffset, &morefragsflag);
        ipdst = 0;
        for (x = 0; x < 4; x++) {
          ipdst <<= 8;
          ipdst |= ipdstarr[x];
        }
        /* check if the destination belongs to my network */
        if ((ipdst & mynetmask) == (myip[0] & mynetmask)) {
          /* compute the dst mac */
          dstmac = arp_getmac(arptable, ipdst);
          if (dstmac == NULL) {
            printf("Could not find the mac for %u.%u.%u.%u in the ARP cache!\n", ipdstarr[0], ipdstarr[1], ipdstarr[2], ipdstarr[3]);
            dstmac = ethbroadcast;
          }
          /* encapsulate the data into an eth frame and relay it */
          forge_eth(&buffptr, &bufflen, mymac, dstmac);
          write(tapfd, buffptr, bufflen);
        }
      }
    }

  }

  return(0);
}
